#include <windows.h>
#include <GL/glew.h>
#include <GL/glut.h>
#include <IL/il.h>
#include <IL/ilu.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <cmath>

#define FOR(__i, __times) for(int __i=0;__i<__times;__i++)

int screenHeight, screenWidth;
bool clicked;

int startX, startY;

#define M_PI 3.1415926535979f

CGcontext context;  
CGprofile fragmentProfile, vertexProfile;
CGprogram fragmentProgram, vertexProgram;

CGparameter eyePosition, env;
CGparameter mv, mvit;

GLuint cubeMap;
float angle0 = 0;
float angle1 = 0;
float angle2 = 0;
float eyeX, eyeY, eyeZ;
float upX, upY, upZ;

GLuint box[6];

double currentRotation[16];
double startRotation[16];

void setIdentity(double m[])
{
	FOR(i,4) FOR(j,4)
		if (i == j)
			m[4*j+i] = 1;
		else
			m[4*j+i] = 0;
}

void mult(double c[], double a[], double b[])
{
	double temp[16];
	FOR(i,16)
		temp[i] = 0;
	FOR(i,4) FOR(j,4) FOR(k,4)
		temp[4*j+i] = a[4*k+i] * b[4*j+k];
	FOR(i,16)
		c[i] = temp[i];
}

void mouse(int button, int state, int x, int y)
{
	if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
	{
		FOR(i,16)
			startRotation[i] = currentRotation[i];
		startX = x;
		startY = y;
		clicked = true;
	}
	else if (button == GLUT_LEFT_BUTTON && state == GLUT_UP)
		clicked = false;
}

void onSphere(double p[], int x, int y)
{
	double xx = 2*(x * 1.0 / screenWidth) - 1;
	double yy = 2*((screenHeight - y) * 1.0 / screenHeight) - 1;
	double ll = sqrt(xx*xx + yy*yy);
	if (ll > 1)
	{
		xx /= ll;
		yy /= ll;
		ll = 1;
	}
	double zz = sqrt(1 - ll*ll);
	
	p[0] = xx; p[1] = yy; p[2] = zz;
}

void cross(double c[], double a[], double b[])
{
	c[0] = a[1]*b[2] - a[2]*b[1];
	c[1] = a[2]*b[0] - a[0]*b[2];
	c[2] = a[0]*b[1] - a[1]*b[0];
}

double length(double a[])
{
	return sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2]);
}

void normalize(double a[])
{
	double ll = length(a);
	a[0] /= ll;
	a[1] /= ll;
	a[2] /= ll;
}

double dot(double a[], double b[])
{
	return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
}

void arcball(int x, int y)
{
	if (x == startX && y == startY)
		return;

	double startVec[3], endVec[3];
	onSphere(startVec, startX, startY);
	onSphere(endVec, x, y);

	double axis[3];
	cross(axis, startVec, endVec);

	double cosTheta = dot(startVec, endVec);
	normalize(axis);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();		
	glRotated(-acos(cosTheta) / M_PI * 180, axis[0], axis[1], axis[2]);
	glMultMatrixd(startRotation);
	glGetDoublev(GL_MODELVIEW_MATRIX, currentRotation);	
}

void updateEyeUp()
{
	eyeX = 10*currentRotation[8] + currentRotation[12];
	eyeY = 10*currentRotation[9] + currentRotation[13];
	eyeZ = 10*currentRotation[10] + currentRotation[14];
	
	upX = currentRotation[4];
	upY = currentRotation[5];
	upZ = currentRotation[6];
}

void motion(int x, int y)
{
	if (clicked)
	{
		arcball(x,y);

		updateEyeUp();

		glutPostRedisplay();	
	}
}

static void checkForCgError(const char *situation)
{
  CGerror error;
  const char *string = cgGetLastErrorString(&error);

  if (error != CG_NO_ERROR) {
    printf("%s: %s\n", situation, string);
    if (error == CG_COMPILER_ERROR) {
      printf("%s\n", cgGetLastListing(context));
    }
    exit(1);
  }
}

float identity[16] = {1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1};

void drawSquare()
{
	glBegin(GL_QUADS);
	glTexCoord2f(1,0);	glVertex2f(-1,-1);
	glTexCoord2f(0,0);	glVertex2f( 1,-1);
	glTexCoord2f(0,1);	glVertex2f( 1, 1);
	glTexCoord2f(1,1);	glVertex2f(-1, 1);
	glEnd();
}

void drawSkyBox()
{
	glEnable(GL_TEXTURE_2D);

	glBindTexture(GL_TEXTURE_2D, box[0]);
	glPushMatrix();
	glScaled(10,10,10);
	glTranslated(1,0,0);
	glRotated(-90, 0, 1, 0);
	
	glBegin(GL_QUADS);
	glTexCoord2f(1,0);	glVertex2f(-1,-1);
	glTexCoord2f(0,0);	glVertex2f( 1,-1);
	glTexCoord2f(0,1);	glVertex2f( 1, 1);
	glTexCoord2f(1,1);	glVertex2f(-1, 1);
	glEnd();

	glPopMatrix();

	glBindTexture(GL_TEXTURE_2D, box[1]);
	glPushMatrix();
	glScaled(10,10,10);
	glTranslated(-1,0,0);
	glRotated(90, 0, 1, 0);
	
	glBegin(GL_QUADS);
	glTexCoord2f(1,0);	glVertex2f(-1,-1);
	glTexCoord2f(0,0);	glVertex2f( 1,-1);
	glTexCoord2f(0,1);	glVertex2f( 1, 1);
	glTexCoord2f(1,1);	glVertex2f(-1, 1);
	glEnd();

	glPopMatrix();

	glBindTexture(GL_TEXTURE_2D, box[2]);
	glPushMatrix();
	glScaled(10,10,10);
	glTranslated(0,1,0);
	glRotated(90, 1, 0, 0);
	
	glBegin(GL_QUADS);
	glTexCoord2f(0,1);	glVertex2f(-1,-1);
	glTexCoord2f(1,1);	glVertex2f( 1,-1);
	glTexCoord2f(1,0);	glVertex2f( 1, 1);
	glTexCoord2f(0,0);	glVertex2f(-1, 1);
	glEnd();

	glPopMatrix();

	glBindTexture(GL_TEXTURE_2D, box[3]);
	glPushMatrix();
	glScaled(10,10,10);
	glTranslated(0,-1,0);
	glRotated(-90, 1, 0, 0);
	
	glBegin(GL_QUADS);
	glTexCoord2f(0,1);	glVertex2f(-1,-1);
	glTexCoord2f(1,1);	glVertex2f( 1,-1);
	glTexCoord2f(1,0);	glVertex2f( 1, 1);
	glTexCoord2f(0,0);	glVertex2f(-1, 1);
	glEnd();

	glPopMatrix();

	glBindTexture(GL_TEXTURE_2D, box[4]);
	glPushMatrix();
	glScaled(10,10,10);
	glTranslated(0,0,1);
	glRotated(180, 0, 1, 0);
	
	glBegin(GL_QUADS);
	glTexCoord2f(1,0);	glVertex2f(-1,-1);
	glTexCoord2f(0,0);	glVertex2f( 1,-1);
	glTexCoord2f(0,1);	glVertex2f( 1, 1);
	glTexCoord2f(1,1);	glVertex2f(-1, 1);
	glEnd();

	glPopMatrix();

	glBindTexture(GL_TEXTURE_2D, box[5]);
	glPushMatrix();
	glScaled(10,10,10);
	glTranslated(0,0,-1);	
	
	glBegin(GL_QUADS);
	glTexCoord2f(1,0);	glVertex2f(-1,-1);
	glTexCoord2f(0,0);	glVertex2f( 1,-1);
	glTexCoord2f(0,1);	glVertex2f( 1, 1);
	glTexCoord2f(1,1);	glVertex2f(-1, 1);
	glEnd();

	glPopMatrix();

	glDisable(GL_TEXTURE_2D);
}

void display()
{
	glClearColor(0,0,0,0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glEnable(GL_CULL_FACE);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60, 1, 0.1, 1000);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(eyeX,eyeY,eyeZ,0,0,0,upX,upY,upZ);

	glEnable(GL_DEPTH_TEST);

	drawSkyBox();	
	
	cgSetMatrixParameterfr(mv, identity);
	cgSetMatrixParameterfr(mvit, identity);
	cgUpdateProgramParameters(vertexProgram);

	cgSetParameter3f(eyePosition, eyeX, eyeY, eyeZ);
	cgGLSetTextureParameter(env, cubeMap);
	cgGLEnableTextureParameter(env);
	cgUpdateProgramParameters(fragmentProgram);	

	cgGLEnableProfile(vertexProfile);
	cgGLEnableProfile(fragmentProfile);
	cgGLBindProgram(vertexProgram);
	cgGLBindProgram(fragmentProgram);

	glutSolidSphere(3, 50, 50);

	cgGLDisableProfile(vertexProfile);
	cgGLDisableProfile(fragmentProfile);

	glutSwapBuffers();
}

void loadImageToTexture(GLuint &id, const char *filename)
{
	glGenTextures(1, &id);
	glBindTexture(GL_TEXTURE_2D, id);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);

	ILuint image;
	ilGenImages(1, &image);
	ilBindImage(image);
	ilLoadImage((wchar_t *)filename);
	iluFlipImage();
	ilConvertImage(IL_RGBA, IL_UNSIGNED_BYTE);

	gluBuild2DMipmaps(GL_TEXTURE_2D, ilGetInteger(IL_IMAGE_BPP), ilGetInteger(IL_IMAGE_WIDTH), 
		ilGetInteger(IL_IMAGE_HEIGHT), ilGetInteger(IL_IMAGE_FORMAT), GL_UNSIGNED_BYTE, ilGetData());	
	
	ilDeleteImages(1, &image);
}

void initEnvironment()
{
	loadImageToTexture(box[0], "../images/cm_right.jpg");
	loadImageToTexture(box[1], "../images/cm_left.jpg");
	loadImageToTexture(box[2], "../images/cm_top.jpg");
	loadImageToTexture(box[3], "../images/cm_bottom.jpg");
	loadImageToTexture(box[4], "../images/cm_front.jpg");
	loadImageToTexture(box[5], "../images/cm_back.jpg");	
}

void loadCubeMapSide(GLuint target, const char *imageName)
{
	ILuint image;
	ilGenImages(1, &image);
	ilBindImage(image);
	ilLoadImage((wchar_t *)imageName);
	ilConvertImage(IL_RGB, IL_UNSIGNED_BYTE);	

	gluBuild2DMipmaps(target, ilGetInteger(IL_IMAGE_BPP), ilGetInteger(IL_IMAGE_WIDTH), 
		ilGetInteger(IL_IMAGE_HEIGHT), ilGetInteger(IL_IMAGE_FORMAT), GL_UNSIGNED_BYTE, ilGetData());

	ilDeleteImages(1, &image);
}

void initCubeMap()
{
	glEnable(GL_TEXTURE_CUBE_MAP_EXT);
	
	glGenTextures(1, &cubeMap);
	glBindTexture(GL_TEXTURE_CUBE_MAP_EXT, cubeMap);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	
	loadCubeMapSide(GL_TEXTURE_CUBE_MAP_POSITIVE_X_EXT, "../images/cm_right.jpg");
	loadCubeMapSide(GL_TEXTURE_CUBE_MAP_NEGATIVE_X_EXT, "../images/cm_left.jpg");
	loadCubeMapSide(GL_TEXTURE_CUBE_MAP_POSITIVE_Y_EXT, "../images/cm_top.jpg");
	loadCubeMapSide(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_EXT, "../images/cm_bottom.jpg");
	loadCubeMapSide(GL_TEXTURE_CUBE_MAP_POSITIVE_Z_EXT, "../images/cm_front.jpg");
	loadCubeMapSide(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_EXT, "../images/cm_back.jpg");
	
	glDisable(GL_TEXTURE_CUBE_MAP_EXT);
}

float m[16];

void animate(int id)
{
	angle0 += 0.4f;
	angle1 += 0.6f;
	angle2 += 0.8f;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glRotated(angle0, 1, 0, 0);
	glRotated(angle1, 0, 1, 0);
	glRotated(angle2, 0, 0, 1);
	glGetFloatv(GL_MODELVIEW_MATRIX, m);

	eyeX = 10*m[8] + m[12];
	eyeY = 10*m[9] + m[13];
	eyeZ = 10*m[10] + m[14];
	
	upX = m[4];
	upY = m[5];
	upZ = m[6];

	glPopMatrix();

	glutPostRedisplay();
	glutTimerFunc(16, animate,0);
}

void initCg()
{
	context = cgCreateContext();
	checkForCgError("creating context");
	cgGLSetDebugMode( CG_FALSE );
	cgSetParameterSettingMode(context, CG_DEFERRED_PARAMETER_SETTING);

	vertexProfile = cgGLGetLatestProfile(CG_GL_VERTEX);
	cgGLSetOptimalOptions(vertexProfile);
	checkForCgError("selecting vertex profile");

	fragmentProfile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
	cgGLSetOptimalOptions(fragmentProfile);
	checkForCgError("selecting fragment profile");

	vertexProgram = cgCreateProgramFromFile(
		context,
		CG_SOURCE,
		"cube-map-vertex.cg",
		vertexProfile,
		"main",
		NULL);
	checkForCgError("creating vertex program");
	cgGLLoadProgram(vertexProgram);
	checkForCgError("loading vertex program");

	mv = cgGetNamedParameter(vertexProgram, "mv");
	checkForCgError("getting mv parameter");
	mvit = cgGetNamedParameter(vertexProgram, "mvit");
	checkForCgError("getting mvit parameter");

	fragmentProgram = cgCreateProgramFromFile(
		context,
		CG_SOURCE,
		"cube-map-fragment.cg",
		fragmentProfile,
		"main",
		NULL);
 	checkForCgError("creating fragment program");
	cgGLLoadProgram(fragmentProgram);
	checkForCgError("loading fragment program");	

	eyePosition = cgGetNamedParameter(fragmentProgram, "eyePosition");
	checkForCgError("getting eyePosition parameter");
	env = cgGetNamedParameter(fragmentProgram, "env");
	checkForCgError("getting env parameter");
}

int main(int argc, char **argv)
{
	ilInit();
	iluInit();
	glewInit();
	glutInit(&argc,  argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	screenWidth = 512;
	screenHeight = 512;
	glutInitWindowSize(screenWidth, screenHeight);
	glutCreateWindow("Cube Map");	
	glutDisplayFunc(display);
	//glutTimerFunc(0, animate, 0);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);

	setIdentity(currentRotation);
	updateEyeUp();

	initCg();	
	initCubeMap();
	initEnvironment();

	glutMainLoop();
}