#include <windows.h>
#include <GL/glew.h>
#include <GL/glut.h>
#include <IL/il.h>
#include <IL/ilu.h>
#include <cmath>

#define M_PI 3.1415926535979f

GLuint cubeMap;
float angle0 = 0;
float angle1 = 0;
float angle2 = 0;

void display()
{
	glClearColor(0,0,0,0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60, 1, 0.1, 1000);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(0,0,5,0,0,0,0,1,0);

	glEnable(GL_TEXTURE_CUBE_MAP_EXT);

	glEnable(GL_DEPTH_TEST);

	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_EXT);
    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_EXT);
    glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_EXT);
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
    glEnable(GL_TEXTURE_GEN_R);
	glEnable(GL_NORMALIZE);

	glutSolidSphere(1.5, 50, 50);
	glRotated(angle0, 1, 0, 0);
	glRotated(angle1, 0, 1, 0);
	glRotated(angle2, 0, 0, 1);
	glutSolidTorus(0.25, 1.75, 50, 50);

	glutSwapBuffers();
}

void loadCubeMapSide(GLuint target, const char *imageName)
{
	ILuint image;
	ilGenImages(1, &image);
	ilBindImage(image);
	ilLoadImage((wchar_t *)imageName);
	ilConvertImage(IL_RGB, IL_UNSIGNED_BYTE);

	gluBuild2DMipmaps(target, ilGetInteger(IL_IMAGE_BPP), ilGetInteger(IL_IMAGE_WIDTH), 
		ilGetInteger(IL_IMAGE_HEIGHT), ilGetInteger(IL_IMAGE_FORMAT), GL_UNSIGNED_BYTE, ilGetData());

	ilDeleteImages(1, &image);
}

void initCubeMap()
{
	glEnable(GL_TEXTURE_CUBE_MAP_EXT);
	glGenTextures(1, &cubeMap);
	glBindTexture(GL_TEXTURE_CUBE_MAP_EXT, cubeMap);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	
	loadCubeMapSide(GL_TEXTURE_CUBE_MAP_POSITIVE_X_EXT, "../images/cm_right.jpg");
	loadCubeMapSide(GL_TEXTURE_CUBE_MAP_NEGATIVE_X_EXT, "../images/cm_left.jpg");
	loadCubeMapSide(GL_TEXTURE_CUBE_MAP_POSITIVE_Y_EXT, "../images/cm_top.jpg");
	loadCubeMapSide(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_EXT, "../images/cm_bottom.jpg");
	loadCubeMapSide(GL_TEXTURE_CUBE_MAP_POSITIVE_Z_EXT, "../images/cm_front.jpg");
	loadCubeMapSide(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_EXT, "../images/cm_back.jpg");
}

void animate(int id)
{
	angle0 += 0.4f;
	angle1 += 0.6f;
	angle2 += 0.8f;
	glutPostRedisplay();
	glutTimerFunc(16, animate,0);
}

int main(int argc, char **argv)
{
	ilInit();
	iluInit();
	glewInit();
	glutInit(&argc,  argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowSize(512, 512);
	glutCreateWindow("Cube Map");	
	glutDisplayFunc(display);
	glutTimerFunc(0, animate, 0);

	initCubeMap();

	glutMainLoop();
}